# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import uuid
from datetime import datetime
from datetime import timezone
from typing import Any
from typing import TypeVar

import pydantic


T = TypeVar('T', bound='MessageMetadata')


class MessageMetadata(pydantic.BaseModel):
    __module__: str = 'aorta.types'
    model_config = {'populate_by_name': True}

    uid: str = pydantic.Field(
        alias='uid',
        default_factory=lambda: str(uuid.uuid4())
    )

    correlation_id: str = pydantic.Field(
        alias='correlationId',
        default_factory=lambda: str(uuid.uuid4())
    )

    published: datetime = pydantic.Field(
        default_factory=lambda: datetime.now(timezone.utc)
    )

    delivery_count: int = pydantic.Field(default=0, alias='deliveryCount')

    ttl: int | None = None

    attempts: int = 0

    handler: str | None = None

    audience: set[str] = set()

    annotations: dict[str, Any] = pydantic.Field(default={})

    labels: dict[str, Any] = pydantic.Field(default={})

    namespace: str = ''

    @pydantic.field_validator('correlation_id', mode='before')
    def preprocess_correlation_id(
        cls,
        value: str | uuid.UUID | None
    ) -> str:
        return cls.preprocess_uuid(value)

    @pydantic.field_validator('uid', mode='before')
    def preprocess_uid(
        cls,
        value: str | uuid.UUID | None
    ) -> str:
        return cls.preprocess_uuid(value)

    @classmethod
    def preprocess_uuid(
        cls,
        value: str | uuid.UUID | None
    ) -> str:
        if value is None:
            value = uuid.uuid4()
        if isinstance(value, uuid.UUID):
            value = str(value)
        uuid.UUID(value)
        return value


# Copyright (C) 2016-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
import fastapi.params

from aorta.types import ITransport
from .aortarouter import AortaRouter
from .localtransport import LocalTransport

__all__: list[str] = [
    'setup_dependencies',
    'AortaRouter',
    'LocalTransport',
]


def setup_dependencies(
    transport: type[ITransport]    
) -> fastapi.params.Depends:
    def f(
        request: fastapi.Request,
        transport: ITransport = fastapi.Depends(transport)
    ):
        setattr(request.state, 'aorta_transport', transport)
    return fastapi.Depends(f)


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import ClassVar
from typing import Self

import pydantic

from .envelope import Envelope


class Message(pydantic.BaseModel):
    __abstract__: bool = True
    __envelope__: type[Envelope[Any]]
    __message_attr__: ClassVar[str]

    def envelope(
        self,
        correlation_id: str | None = None,
        audience: set[str] | None = None,
        labels: dict[str, str | None] | None = None,
        annotations: dict[str, str | None] | None = None,
        namespace: str = ''
    ) -> Envelope[Self]:
        return self.__envelope__.model_validate({
            'apiVersion': getattr(self, '__version__', 'v1'),
            'kind': type(self).__name__,
            'type': 'unimatrixone.io/command',
            'metadata': {
                'annotations': annotations or {},
                'audience': audience or set(),
                'correlationId': correlation_id,
                'labels': labels or {},
                'namespace': namespace
            },
            self.__message_attr__: self.model_dump()
        })


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol

from .envelope import Envelope


class Publishable(Protocol): # pragma: no cover
    __module__: str = 'aorta.types'

    def envelope(
        self,
        correlation_id: str | None = None,
        audience: set[str] | None = None
    ) -> Envelope[Any]:
        ...


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import ClassVar
from typing import Self

from .droppable import Droppable
from .envelope import Envelope


class EnvelopeMixin(Droppable):
    __abstract__: ClassVar[bool] = True
    __envelope__: ClassVar[type[Envelope[Self]]]
    __message_attr__: ClassVar[str]
    __message_type__: ClassVar[str]

    def envelope(
        self,
        correlation_id: str | None = None,
        audience: set[str] | None = None,
        labels: dict[str, str | None] | None = None,
        annotations: dict[str, str | None] | None = None,
        namespace: str = '',
    ) -> Envelope[Self]:
        return self.__envelope__.model_validate({
            'apiVersion': getattr(self, '__version__', 'v1'),
            'kind': type(self).__name__,
            'type': self.__message_type__,
            'metadata': {
                'annotations': annotations or {},
                'audience': audience or set(),
                'correlationId': correlation_id,
                'labels': labels or {},
                'namespace': namespace
            },
            self.__message_attr__: self.model_dump() # type: ignore
        })


# Copyright (C) 2020-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import threading
from typing import Any
from typing import Iterable

import fastapi

from aorta.types import Acknowledgable
from aorta.types import IPublisher
from aorta.types import IMessageHandler
from aorta.types import Envelope
from ._base import BaseRunner


class SequentialRunner(BaseRunner):
    """Runs all message handlers sequentially."""

    async def run(
        self,
        publisher: IPublisher,
        frame: Acknowledgable,
        envelope: Envelope[Any],
        handlers: Iterable[type[IMessageHandler]],
        request: fastapi.Request | None = None
    ) -> None:
        t = threading.current_thread()
        self.logger.debug(
            "Running message (kind: %s, uid: %s, correlation-id: %s, thread: %s)",
            envelope.kind,
            envelope.metadata.uid,
            envelope.metadata.correlation_id,
            t.native_id
        )
        tasks: list[asyncio.Task[Any]] = []
        handlers = list(handlers)
        for handler_class in handlers:
            c = self.run_handler(
                publisher=publisher,
                frame=frame,
                envelope=envelope,
                handler_class=handler_class,
                request=request
            )
            t = asyncio.create_task(c)
            tasks.append(t)
        await asyncio.gather(*tasks, return_exceptions=True)


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
import uuid
from typing import Any

from .types import Envelope
from .types import IPublisher
from .types import MessageMetadata
from .types import Publishable


class Transaction:
    __module__: str = 'aorta'
    correlation_id: str
    logger: logging.Logger = logging.getLogger('canonical')
    messages: list[Envelope[Any]]
    publisher: IPublisher

    def __init__(
        self,
        publisher: IPublisher,
        metadata: MessageMetadata | None = None,
        correlation_id: str | None = None
    ):
        self.correlation_id = correlation_id or str(uuid.uuid4())
        self.messages = []
        self.metadata = metadata
        self.publisher = publisher

    def issue(
        self,
        message: Publishable,
        correlation_id: str | None = None,
        audience: set[str] | None = None
    ):
        return self.publish(message, correlation_id=correlation_id, audience=audience)

    def get_correlation_id(self) -> str:
        cid = self.correlation_id
        if self.metadata:
            cid = self.metadata.correlation_id
        return cid

    def pending(self) -> list[Publishable]:
        return [x.message for x in self.messages]

    def publish(
        self,
        message: Publishable,
        correlation_id: str | None = None,
        audience: set[str] | None = None
    ):
        envelope = message.envelope(
            correlation_id=correlation_id or self.correlation_id,
            audience=audience
        )
        self.logger.debug(
            "Adding %s to running transaction (uid: %s, correlation-id: %s)",
            type(message).__name__,
            envelope.metadata.uid,
            envelope.metadata.correlation_id
        )
        self.messages.append(envelope)

    def rollback(self) -> None:
        self.messages = []

    async def send(self, message: Publishable) -> None:
        envelope = message.envelope(correlation_id=self.get_correlation_id())
        await self.publisher.send([envelope])

    async def commit(self) -> None:
        """Immediately publishes pending messages to the messaging
        infrastructure.
        """
        self.logger.debug(
            "Committed %s messages (uid: %s, correlation-id: %s)",
            len(self.messages),
            self.metadata.uid if self.metadata else 'None',
            self.correlation_id
        )
        await self.publisher.send(self.messages)
        self.messages = []

    async def __aenter__(self):
        return self
    
    async def __aexit__(self, cls: type[BaseException] | None, *args: Any):
        if cls is not None:
            self.rollback()
            return
        await self.commit()


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from datetime import timezone

import pydantic

from .messagemetadata import MessageMetadata


class MessageHeader(pydantic.BaseModel):
    __module__: str = 'aorta.types'
    api_version: str = pydantic.Field(..., alias='apiVersion')
    kind: str
    type: str
    metadata: MessageMetadata = pydantic.Field(
        default_factory=MessageMetadata
    )

    @property
    def age(self) -> int:
        """Return the age of the message as a UNIX timestamp, in milliseconds
        since the UNIX epoch.
        """
        return int((datetime.now(timezone.utc) - self.metadata.published).total_seconds())

    @property
    def qualname(self) -> tuple[str, str]:
        """Return the qualified name of the message type."""
        return (self.api_version, self.kind)

    def accept(self):
        """Accepts the message. This method is invoked by the runner of the
        message handlers.
        """
        self.metadata.delivery_count += 1

    def is_command(self) -> bool:
        """Return a boolean indicating if the message is a command."""
        return self.type == "cochise.io/command"

    def is_event(self) -> bool:
        """Return a boolean indicating if the message is an event."""
        return self.type == "cochise.io/event"

    def is_expired(self) -> bool:
        """Return a boolean indicating if the message is expired."""
        return (self.age > self.metadata.ttl)\
            if self.metadata.ttl\
            else False

    def is_known(self) -> bool:
        """Return a boolean if the enclosed message is known."""
        return False

    def is_private_event(self) -> bool:
        """Return a boolean indicating if the enclosed message is a
        private event.
        """
        return False

    def is_valid(self, now: int | None = None) -> bool:
        """Return a boolean indicating if the message is still valid."""
        return not self.is_expired()

    def __bytes__(self) -> bytes:
        return str.encode(
            self.model_dump_json(by_alias=True, exclude_defaults=True),
            "utf-8"
        )



# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable

from .types import Command
from .types import Event
from .types import Envelope
from .types import IPublisher
from .types import ITransaction
from .types import ITransport
from .types import Message
from .transaction import Transaction


class MessagePublisher(IPublisher):
    __module__: str = 'aorta'
    transport: ITransport
    transaction_factory: Callable[[IPublisher], ITransaction]

    def __init__(
        self,
        transport: ITransport,
        transaction_factory: Callable[[IPublisher], ITransaction] | None = None
    ):
        self.transport = transport
        self.transaction_factory = transaction_factory or Transaction

    def begin(self) -> ITransaction:
        return self.transaction_factory(self)

    async def publish(
        self,
        message: Message | Command | Event,
        **kwargs: Any
    ) -> None:
        await self.send([message.envelope(**kwargs)])

    async def send(self, messages: list[Envelope[Any]], is_retry: bool = False) -> None:
        return await self.transport.send(messages, is_retry=is_retry)


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from .command import Command
from .event import Event
from .itransaction import ITransaction
from .messagemetadata import MessageMetadata
from .publishable import Publishable


class IMessageHandler:
    __module__: str = 'aorta.types'

    @property
    def qualname(self) -> str:
        ...

    def __init__(
        self,
        publisher: ITransaction,
        metadata: MessageMetadata
    ):
        ...

    def issue(self, command: Command, audience: set[str] | None = None) -> None: ...
    def publish(self, event: Event, audience: set[str] | None = None) -> None: ...
    async def commit(self) -> None: ...
    async def send(self, message: Publishable) -> None: ...
    async def run(self, envelope: Any) -> tuple[bool, Any]: ...


# Copyright (C) 2016-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import threading
import time
from typing import Any
from typing import Callable
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures.thread import BrokenThreadPool
from queue import Empty
from queue import SimpleQueue

import aorta
from aorta.runners import SequentialRunner
from aorta.types import Acknowledgable
from aorta.types import Envelope
from aorta.types import IRunner
from aorta.types import IPublisher
from ._pool import Pool


class ThreadPool(Pool):
    must_exit: bool = False
    step: int = 0

    def __init__(
        self,
        publisher: IPublisher,
        queue: SimpleQueue[tuple[Acknowledgable, Envelope[Any]]],
        concurrency: int = 1,
        debug: bool = False,
        runner: IRunner | None = None,
        initializer: Callable[..., None] | None = None
    ) -> None:
        super().__init__(queue=queue, concurrency=concurrency, debug=debug)
        self.initializer = initializer or (lambda: None)
        self.executor = ThreadPoolExecutor(
            max_workers=concurrency,
            initializer=self.initializer
        )
        self.publisher = publisher
        self.thread = threading.Thread(
            daemon=True,
            target=self.main_event_loop
        )
        self.runner = runner or SequentialRunner()
        self.thread.start()

    def main_event_loop(self):
        self.logger.debug("Starting pool manager thread.")
        while True:
            time.sleep(0.1)
            self.step += 1
            self.main_event()
            if self.must_exit:
                break

    def main_event(self) -> None:
        try:
            frame, envelope = self.queue.get(block=True, timeout=1.0)
        except Empty:
            return
        try:
            self.executor.submit(
                self.run,
                ack=lambda x: x.ack(),
                runner=self.runner,
                publisher=self.publisher,
                frame=frame,
                envelope=envelope
            )
        except BrokenThreadPool:
            self.logger.error("Pool is broken, reinitializing.")
            self.executor.shutdown(wait=False, cancel_futures=False)
            self.executor = ThreadPoolExecutor(
                max_workers=self.concurrency,
                initializer=self.initializer
            )
            self.executor.submit(
                self.run,
                ack=lambda x: x.ack(),
                runner=self.runner,
                publisher=self.publisher,
                frame=frame,
                envelope=envelope
            )
        except RuntimeError:
            # This is raised if shutdown() was called.
            pass
        except Exception as e:
            self.logger.exception(
                "Caught fatal %s (uid: %s, correlation-id: %s)",
                type(e).__name__,
                envelope.metadata.uid,
                envelope.metadata.correlation_id,
            )

    def run(
        self,
        ack: Callable[[Acknowledgable], None],
        runner: IRunner,
        publisher: IPublisher,
        frame: Acknowledgable,
        envelope: Envelope[Any],
    ) -> None:
        loop = asyncio.new_event_loop()
        try:
            c = runner.run(
                publisher=publisher,
                frame=frame,
                envelope=envelope,
                handlers=aorta.get(envelope)
            )
            loop.run_until_complete(c)
        finally:
            ack(frame)
            self.logger.debug(
                "Acknowledged frame (uid: %s, correlation-id: %s)",
                envelope.metadata.uid,
                envelope.metadata.correlation_id
            )
            if loop.is_running():
                loop.close()

    async def join(self):
        self.must_exit = True
        self.executor.shutdown(wait=True, cancel_futures=False)


# Copyright (C) 2016-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging.config
import os
from typing import Any
from typing import TypeVar
from queue import SimpleQueue

import fastapi
from libcanonical.runtime import MainProcess # type: ignore

from aorta import Command
from aorta import Event
from aorta import MessagePublisher
from aorta import NullTransport
from aorta.types import Acknowledgable
from aorta.types import Envelope
from aorta.types import IPublisher
from aorta.types import IRunner
from ._threadpool import ThreadPool
from ._pool import Pool


E = TypeVar('E', bound=Command|Event)


class BaseConsumer(MainProcess):
    __module__: str = 'aorta.ext.consumer'
    concurrency: int = 1
    debug: bool = False
    interval = 0.01
    pool: Pool
    queue: SimpleQueue[tuple[Acknowledgable, Envelope[Any]]]

    def __init__(self, name: str, concurrency: int = 1, **kwargs: Any):
        super().__init__(name=name)
        self.concurrency = concurrency
        if os.getenv('WORKER_CONCURRENCY') and str.isdigit(os.environ['WORKER_CONCURRENCY']):
            self.concurrency = int(os.environ['WORKER_CONCURRENCY'])

    def acknowledge(self, message: Acknowledgable) -> None:
        raise NotImplementedError

    def configure(self, reloading: bool = False):
        self.logger.debug("Running with %s workers", self.concurrency)
        if not reloading:
            self.publisher = self.get_publisher()
            self.queue = SimpleQueue()
            self.pool = ThreadPool(
                publisher=self.publisher,
                queue=self.queue,
                concurrency=self.concurrency,
                debug=self.debug,
                initializer=self.configure_worker
            )

    def configure_worker(self) -> None:
        logging.config.dictConfig(dict(self.get_logging_config()))

    def get_publisher(self) -> IPublisher:
        return MessagePublisher(
            transport=NullTransport()
        )

    def get_runner(
        self,
        publisher: IPublisher,
        frame: Acknowledgable,
        request: fastapi.Request | None = None,
        **kwargs: Any
    ) -> IRunner:
        raise NotImplementedError

    def on_message(
        self,
        frame: Acknowledgable,
        envelope: Envelope[Any]
    ) -> None:
        raise NotImplementedError

    def route(self, envelope: Envelope[E]) -> list[str]:
        """Return a list of strings indicating the topics that the
        `envelope` must be sent to. Subclasses must override this
        method.
        """
        raise NotImplementedError

    async def main_event(self) -> None:
        pass

    async def teardown(self) -> None:
        await self.pool.join()

    def _acknowledge(self, message: Acknowledgable, envelope: Envelope[Any]) -> None:
        try:
            self.acknowledge(message)
        except Exception:
            self.logger.exception(
                "Unable to ACK message (uid: %s, correlatio-id: %s)",
                envelope.metadata.uid,
                envelope.metadata.correlation_id,
            )


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import TypeVar

import pydantic

from .envelopedmixin import EnvelopeMixin
from .eventtype import EventType


T = TypeVar('T', bound='Event')


class Event(pydantic.BaseModel, EnvelopeMixin, metaclass=EventType):
    __module__: str = 'aorta.types'
    __message_attr__ = 'data'
    __message_type__ = 'cochise.io/event'


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic
import pydantic.main

from .envelope import Envelope


class MessageType(type(pydantic.BaseModel)):
    __module__: str = 'aorta.types'
    __registry__: dict[tuple[str, str], type[Any]]
    envelope_attr: str
    envelope_class: type[Envelope[Any]]
    typename: str

    def __new__(
        cls,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **params: Any
    ) -> 'MessageType':
        is_abstract = namespace.pop('__abstract__', False)
        skip_register: bool = namespace.pop('__aorta_disable__', False)
        new_class =  super().__new__(cls, name, bases, namespace, **params) # type: ignore
        namespace.setdefault('__version__', 'v1')
        if not is_abstract:
            k: tuple[str, str] = (namespace['__version__'], name)
            if k in cls.__registry__:
                raise TypeError('Message {0}/{1} is already registered.'.format(*k))
            new_class.__envelope__ = type(
                f'{name}Envelope',
                (cls.envelope_class,), # type: ignore
                {
                    '__annotations__': {
                        cls.envelope_attr: new_class
                    }
                }
            )
            if not skip_register:
                cls.__registry__[k] = new_class.__envelope__ # type: ignore
        
        return new_class # type: ignore


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
from typing import Any
from typing import Callable

from google.cloud import pubsub_v1 # type: ignore

from aorta.types import Envelope


class PubsubTransport:
    __module__: str = 'aorta.ext.google'
    client: pubsub_v1.PublisherClient
    retry_topic: str | None
    opic: str | list[str] | Callable[..., list[str]]

    def __init__(
        self,
        *,
        project: str,
        topic: str | list[str] | Callable[..., list[str]],
        retry_topic: str | None = None
    ):
        """Initialize a new :class:`GoogleTransport`:

        Args:
            project: the name of the Google Cloud project.
            topic_path: either a string, list of string, or a callable that
                returns a string or list of strings, that specify the topic
                to which messages must be published.
        """
        self.client = pubsub_v1.PublisherClient()
        self.project = project
        self.retry_topic = retry_topic
        self.topic = topic

    def get_topics(self, envelope: Envelope[Any], is_retry: bool = False) -> list[str]:
        """Return the list of topics to which the given `message` must be
        published.
        """
        if is_retry:
            topics = self.retry_topic
        elif callable(self.topic):
            topics = self.topic(envelope)
        else:
            topics = self.topic
        assert isinstance(topics, (str, list)) # nosec
        topics = [topics] if isinstance(topics, str) else topics
        return [self.client.topic_path(self.project, x) for x in topics]

    async def send(
        self,
        messages: list[Envelope[Any]],
        is_retry: bool = False
    ) -> None:
        futures: list[asyncio.Future[Any]] = []
        for envelope in messages:
            for topic in self.get_topics(envelope, is_retry=is_retry):
                futures.append(
                    asyncio.ensure_future(self._send(topic, envelope))
                )
        await asyncio.gather(*futures)

    async def _send(self, topic: str, message: Envelope[Any]):
        future: asyncio.Future[Any] = asyncio.wrap_future(
            future=self.client.publish(topic, bytes(message)) # type: ignore
        )
        await future